# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""All command-line functionality."""

from . import (
    annotate,
    convert_dep5,
    download,
    lint,
    lint_file,
    main,
    spdx,
    supported_licenses,
)
