# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Entry module for reuse."""

if __name__ == "__main__":
    from .cli.main import main

    # pylint: disable=no-value-for-parameter
    main()
